'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault(ex) {
  return ex && typeof ex === 'object' && 'default' in ex ? ex['default'] : ex;
}

var pupa = _interopDefault(require('pupa'));
var isNumber = _interopDefault(require('is-number'));
var get = _interopDefault(require('dot-prop-wild'));

var messages = {
  accepted: 'This field must be accepted',
  alpha: 'This field may only contain letters',
  alphaNum: 'This field may only contain letters and numbers',
  alphaNumDash: 'This field may only contain letters, numbers, dashes and underscores',
  array: 'This field must be an array',
  between: {
    array: 'This field must have between {min} and {max} items',
    numeric: 'This field must be between {min} and {max}',
    string: 'This field must be between {min} and {max} characters'
  },
  boolean: 'This field must be true or false',
  email: 'This field must be a valid email address',
  max: {
    array: 'This field may not have more than {max} items',
    numeric: 'This field may not be greater than {max}',
    string: 'This field may not be greater than {max} characters'
  },
  min: {
    array: 'This field must have at least {min} items',
    numeric: 'This field must be at least {min}',
    string: 'This field must be at least {min} characters'
  },
  numeric: 'This field must be a number',
  required: 'This field is required',
  slug: 'This field may only contain lowercase letters, numbers and dashes',
  string: 'This field must be a string'
};

function addMessage(title, message) {
  messages[title] = message;
}

function formatErrorMessage(rule, params = {}, rules) {
  const errorMessage = messages[rule];

  if (!errorMessage) {
    throw new Error(`There is no defined error message for '${rule}' rule`);
  }

  if (!rules) {
    throw new Error('All rules were not passed');
  }

  if (typeof errorMessage === 'object') {
    const ruleValueType = getRuleValueType(rules);

    return pupa(errorMessage[ruleValueType], params);
  }

  return pupa(errorMessage, params);
}

var accepted = () =>
  function accepted(value) {
    return {
      valid: value === 'yes' || value === 'on' || value === 1 || value === true
    };
  };

function isString(value) {
  return typeof value === 'string';
}

var alpha = () =>
  function alpha(value) {
    return {
      valid: isString(value) && /^$|^[a-z]+$/i.test(value)
    };
  };

var alphaNum = () =>
  function alphaNum(value) {
    return {
      valid: (isString(value) || isNumber(value)) && /^$|^[a-z0-9]+$/i.test(value)
    };
  };

var alphaNumDash = () =>
  function alphaNumDash(value) {
    return {
      valid: (isString(value) || isNumber(value)) && /^$|^[a-z0-9_-]+$/i.test(value)
    };
  };

var array = () =>
  function array(value) {
    return {
      valid: Array.isArray(value)
    };
  };

var between = (min, max) => {
  if (!min) {
    throw new Error('between rule "min" param was not set');
  }

  if (!max) {
    throw new Error('between rule "max" param was not set');
  }

  return function between(value, rules) {
    const valueSize = getValueSize(value, rules);

    return {
      valid: valueSize >= min && valueSize <= max,
      params: { min, max }
    };
  };
};

var boolean = () =>
  function boolean(value) {
    return {
      valid: value === true || value === false
    };
  };

var email = () =>
  function email(value) {
    return {
      valid: isString(value) && /^$|^.+@.+$/.test(value)
    };
  };

var max = param => {
  if (!param) {
    throw new Error('max rule param was not set');
  }

  return function max(value, rules) {
    const valueSize = getValueSize(value, rules);

    return {
      valid: valueSize <= param,
      params: { max: param }
    };
  };
};

var min = param => {
  if (!param) {
    throw new Error('min rule param was not set');
  }

  return function min(value, rules) {
    const valueSize = getValueSize(value, rules);

    return {
      valid: valueSize >= param,
      params: { min: param }
    };
  };
};

var numeric = () =>
  function numeric(value) {
    return {
      valid: isNumber(value)
    };
  };

var required = () =>
  function required(value) {
    if (value === undefined || value === null) {
      return {
        valid: false
      };
    }

    if (Array.isArray(value) && !value.length) {
      return {
        valid: false
      };
    }

    const convertedString = String(value).replace(/\s/g, '');

    return {
      valid: convertedString.length > 0
    };
  };

var slug = () =>
  function slug(value) {
    return {
      valid: isString(value) && /^$|^[a-z0-9]+(?:-[a-z0-9]+)*$/.test(value)
    };
  };

var string = () =>
  function string(value) {
    return {
      valid: isString(value)
    };
  };

const validationRules = {
  accepted,
  alpha,
  alphaNum,
  alphaNumDash,
  array,
  between,
  boolean,
  email,
  max,
  min,
  numeric,
  required,
  slug,
  string
};

const numericRules = ['numeric'];
const mandatoryRules = ['accepted', 'required'];

function addRule(title, ruleFunction, message) {
  if (!title || !ruleFunction || !message) {
    throw new Error('No title, function or message was provided for custom rule');
  }

  if (validationRules[title]) {
    throw new Error(`Validation rule '${title}' already exists`);
  }

  validationRules[title] = ruleFunction;

  addMessage(title, message);
}

function mandatoryRule(rule, value) {
  if (mandatoryRules.includes(rule.name)) {
    return true;
  }

  return value !== undefined;
}

function rulesContain(rules, rulesToFind) {
  if (Array.isArray(rulesToFind)) {
    let ruleWasFound = false;

    for (let i = 0; i < rulesToFind.length; i += 1) {
      if (rules.includes(rulesToFind[i])) {
        ruleWasFound = true;

        break;
      }
    }

    return ruleWasFound;
  }

  return rules.includes(rulesToFind);
}

function rulesContainNumericRules(rules) {
  return rulesContain(rules, numericRules);
}

function getRuleValueType(rules) {
  if (rulesContainNumericRules(rules)) {
    return 'numeric';
  }

  if (rulesContain(rules, 'array')) {
    return 'array';
  }

  return 'string';
}

function getValueSize(value, rules) {
  if (isNumber(value) && rulesContainNumericRules(rules)) {
    return Number(value);
  }

  if (Array.isArray(value)) {
    return value.length;
  }

  return String(value).trim().length;
}

async function validateValue({ value }, rules) {
  const ruleNames = rules.map(rule => rule.name);
  let errorMessage;

  for (let i = 0; i < rules.length; i += 1) {
    const rule = rules[i];

    if (mandatoryRule(rule, value)) {
      const { valid, params } = await rule(value, ruleNames);

      if (!valid) {
        errorMessage = formatErrorMessage(rule.name, params, ruleNames);

        break;
      }
    }
  }

  return errorMessage;
}

async function validatePath(path, rules, data) {
  const value = get(data, path);
  const error = await validateValue(value, rules);

  return error ? { [value.path]: error } : null;
}

async function validateWildcardPath(path, rules, data) {
  const values = get(data, path);
  const errors = await Promise.all(
    values.map(async value => {
      const error = await validateValue(value, rules);

      return error ? { [value.path]: error } : null;
    })
  );
  const filteredErrors = errors.filter(error => error);

  return filteredErrors.length > 0 ? Object.assign({}, ...filteredErrors) : null;
}

async function validateData(data, rules) {
  const rulesArray = Object.keys(rules);

  if (!rulesArray.length) {
    return [];
  }

  const errors = await Promise.all(
    rulesArray.map(async path => {
      const pathRules = rules[path];

      return !path.includes('*')
        ? validatePath(path, pathRules, data)
        : validateWildcardPath(path, pathRules, data);
    })
  );

  return errors.filter(error => error);
}

async function validate(data, rules) {
  if (!data) {
    throw new Error('No data for validation was provided');
  }

  if (!rules) {
    throw new Error('No validation rules were provided');
  }

  const errors = await validateData(data, rules);

  return {
    valid: errors.length === 0,
    errors: Object.assign({}, ...errors)
  };
}

exports.addRule = addRule;
exports.rules = validationRules;
exports.validate = validate;
